/* define to the path to cpp */
#undef CPP_PATH

/* package name */
#undef PACKAGE 

/* package version */
#undef VERSION 

/* define if you want GNOME stuff support */
#undef GNOME_STUFF

/* define if you want KDE hint support */
#undef KWM_HINTS
